from typing import Callable

from .syntax import (
    CODE_REGEX,
    HEAD_REGEX,
    HTML_REGEX,
    LINK_REGEX,
    LIST_REGEX,
    MASK_SYMBOL,
    NULL_SYMBOL,
    QUOTE_REGEX,
    SYMBOL_REGEX,
    URL_REGEX,
)


def fill(symbol: str) -> Callable:
    def _runner(matched) -> str:
        return symbol * len(matched[0])

    return _runner


def clean(line):
    """
    * Remove URL
    * Remove Markdown symbols and fill by spaces.
        * List * - 1. * [ ] ...
        * Quote > ...
        * Heading # ...
        * Link and Image [...]()
        * Bold * _
        * Strikethrough ~
    * Mask texts inside Codes `...`
    """
    line.body = LIST_REGEX.sub(fill(" "), line.body)
    line.body = QUOTE_REGEX.sub(fill(" "), line.body)
    line.body = HEAD_REGEX.sub(fill(" "), line.body)
    line.body = LINK_REGEX.sub(
        lambda m: NULL_SYMBOL * len(m[1]) + m[2] + NULL_SYMBOL * len(m[3]), line.body
    )
    line.body = HTML_REGEX.sub(fill(NULL_SYMBOL), line.body)
    line.body = URL_REGEX.sub(fill(NULL_SYMBOL), line.body)
    line.body = SYMBOL_REGEX.sub(fill(NULL_SYMBOL), line.body)
    line.body = CODE_REGEX.sub(fill(MASK_SYMBOL), line.body)
    return line


def clean_lines(lines):
    for line in lines:
        if not line.is_ignore:
            clean(line)
        yield line



"""
1. Split body line by line
2. Detect Japanese inside each lines
3. Detect line types
    * Header
    * List
    * General
4. Clean up Japanese lines
    * Remove URLs
    * Remove ASCII symbols
5. Detect Paragraphs and wrap by models
"""
from typing import Iterable

from .cleaner import clean_lines
from .lines import parse_lines
from .paragraph import Paragraph, parse_paragraph


def parse(body: str) -> Iterable[Paragraph]:
    lines = parse_lines(body)
    lines = clean_lines(lines)
    return parse_paragraph(lines)



import re

NULL_SYMBOL = "~"
MASK_SYMBOL = "_"

JA_REGEX = re.compile(
    "["
    "\u3041-\u309F"  # HIRAGANA
    "\u30A1-\u30FF"  # KATAKANA
    "\u2E80-\u2FDF\u3005-\u3007\u3400-\u4DBF\u4E00-\u9FFF\uF900-\uFAFF"  # KANJI
    "\U00020000-\U0002EBEF"  # KANJI
    "]+"
)

HEAD_REGEX = re.compile("^#+")
LIST_REGEX = re.compile(r"^\s*([*-]|\d+\.) (\[[x ]?] )?")
QUOTE_REGEX = re.compile(r"^>+\s*")
HTML_REGEX = re.compile(r"<.+?>")

URL_REGEX = re.compile(r"<?(http|https|ftp)://[-a-zA-Z0-9()@:%_+.~#?&/=]+>?")
LINK_REGEX = re.compile(r"(!?\[)(.*?)(]\(.*?\))")
CODE_REGEX = re.compile(r"`.+?`")
SYMBOL_REGEX = re.compile(r"(\*|___|__|~)")  # ASCII Symbols to clean

FENCE_START = re.compile("^```.+$")
FENCE_END = re.compile("^```$")



from dataclasses import dataclass
from typing import Iterable, List, Optional

from .lines import Line
from .syntax import NULL_SYMBOL


@dataclass
class Pos:
    """Line number and column position.
    These numbers starts from Zero.
    """

    line: int
    ch: int

    def __eq__(self, other) -> bool:
        return self.line == other.line and self.ch == other.ch

    def __repr__(self) -> str:
        return f"Pos({self.line}, {self.ch})"


@dataclass
class Paragraph:
    number: int
    index: int
    lines: List[str]
    is_head: bool = False
    is_list: bool = False
    _body: Optional[str] = None

    def __eq__(self, other):
        return (
            self.number == other.number
            and self.index == other.index
            and self.lines == other.lines
            and self.is_head is other.is_head
            and self.is_list is other.is_list
        )

    @property
    def body(self):
        if self._body is not None:
            return self._body
        self._body = "\n".join(line.strip().replace(NULL_SYMBOL, "") for line in self.lines)
        return self._body

    def as_original_pos(self, index) -> Pos:
        before_lines = self.body[:index].split("\n")
        pos = Pos(
            len(before_lines) - 1,
            len(before_lines[-1]),
        )
        # Convert column number to original position.
        line = self.lines[pos.line]
        left_margin = len(line) - len(line.lstrip())
        chunks = line[left_margin:].split(NULL_SYMBOL)
        num_null_symbols = 0
        current = 0
        for num_null_symbols, chunk in enumerate(chunks):
            current += len(chunk)
            if pos.ch < current:
                break
        pos.line += self.number
        pos.ch += left_margin + num_null_symbols
        return pos

    def as_original_index(self, index: int) -> int:
        pos = self.as_original_pos(index)
        num_lines = pos.line - self.number
        before_index = sum(len(line) for line in self.lines[:num_lines]) + num_lines
        return self.index + before_index + pos.ch

    def __str__(self) -> str:
        return self.body

    def __repr__(self) -> str:
        body = self.body
        if self.is_head:
            body = "# " + body
        elif self.is_list:
            body = "* " + body
        return f"Paragraph({self.number}. ({self.index}) {body})"

    def append(self, line: str):
        self.lines.append(line)

    def __bool__(self) -> bool:
        return bool(self.lines)


def parse_paragraph(lines: Iterable[Line]) -> Iterable[Paragraph]:
    """
    Combine multiple lines into one Paragraph.
    Each paragraphs has chunks.
    """
    paragraph = Paragraph(0, 0, [])
    index = 0
    for i, line in enumerate(lines):
        next_index = index + len(line.body) + 1

        if line.is_ignore or line.is_head or line.is_list:
            if paragraph:
                yield paragraph
            paragraph = Paragraph(i + 1, next_index, [])

        if line.is_head or line.is_list:
            # Yield my self as 1 line paragraph.
            yield Paragraph(
                i,
                index,
                [line.body],
                is_head=line.is_head,
                is_list=line.is_list,
            )
        elif not line.is_ignore:
            paragraph.append(line.body)

        index = next_index

    if paragraph:
        yield paragraph
